"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const zlib_1 = require("zlib");
const AbcFileReader_1 = require("./abcFile/AbcFileReader");
const SmarterBuffer_1 = require("./SmarterBuffer");
const Types_1 = require("./Types");
/**
 * Spec: https://wwwimages2.adobe.com/content/dam/acom/en/devnet/pdf/swf-file-format-spec.pdf
 */
class SwfReader {
    constructor(path) {
        this.path = path;
        this.readPath();
    }
    getFile() {
        const header = this.getHeader();
        const tags = this.getTags();
        return { header, tags };
    }
    getHeader() {
        this.readPath();
        const signatureString = String.fromCharCode(...[
            this.buffer.readUInt8(), this.buffer.readUInt8(), this.buffer.readUInt8(),
        ]);
        const signature = signatureString;
        const version = this.buffer.readUInt8();
        const fileLength = this.buffer.readUInt32LE();
        this.buffer = SmarterBuffer_1.default.fromBuffer(signature === Types_1.SwfSignature.Zlib
            ? zlib_1.inflateSync(this.buffer.readBuffer())
            : this.buffer.readBuffer());
        const frameSize = this.buffer.readRect();
        const frameRate = this.buffer.readFixed(16);
        const frameCount = this.buffer.readUInt16LE();
        return { signature, version, fileLength, frameSize, frameRate, frameCount };
    }
    getTags() {
        const tags = [];
        let tag;
        do {
            tag = this.readTag();
            tags.push(tag);
        } while (tag.code !== Types_1.TagCode.End);
        return tags;
    }
    readTag() {
        const tagCodeAndLength = this.buffer.readUInt16LE();
        const code = tagCodeAndLength >> 6;
        let length = tagCodeAndLength & 63;
        if (length === 63) {
            length = this.buffer.readUInt32LE();
        }
        const data = this.buffer.readBuffer(length);
        switch (code) {
            /**
             * A partir de la page 18 de la spec
             * https://wwwimages2.adobe.com/content/dam/acom/en/devnet/pdf/avm2overview.pdf
             */
            case Types_1.TagCode.DoABC:
                const dataBuffer = SmarterBuffer_1.default.fromBuffer(data);
                const flags = dataBuffer.readUInt32LE();
                const name = dataBuffer.readStringNT();
                const abcFileReader = new AbcFileReader_1.default(dataBuffer);
                return {
                    abcFile: abcFileReader.readFile(),
                    code,
                    flags,
                    length,
                    name,
                };
            case Types_1.TagCode.End:
                return { code, length };
            default:
                return { code, length, data };
        }
    }
    readPath() {
        this.buffer = SmarterBuffer_1.default.fromBuffer(fs_1.readFileSync(this.path));
    }
}
exports.default = SwfReader;
