"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const bytecode_1 = require("./types/bytecode");
const constant_1 = require("./types/constant");
const instance_1 = require("./types/instance");
const methods_1 = require("./types/methods");
const multiname_1 = require("./types/multiname");
const trait_1 = require("./types/trait");
/**
 * Spec: https://wwwimages2.adobe.com/content/dam/acom/en/devnet/pdf/avm2overview.pdf
 * Page 18 onwards
 */
class AbcFileReader {
    constructor(buffer) {
        this.buffer = buffer;
    }
    readFile() {
        const minorVersion = this.buffer.readUInt16LE();
        const majorVersion = this.buffer.readUInt16LE();
        const constantPool = this.readConstantPool();
        const methodCount = this.buffer.readEncodedU30();
        const methods = this.readMethods(methodCount, constantPool);
        const metadataCount = this.buffer.readEncodedU30();
        const metadataInfos = this.readMetadataInfos(metadataCount, constantPool);
        const classCount = this.buffer.readEncodedU30();
        const classes = [];
        const methodBodies = [];
        const instances = this.readInstances(classCount, constantPool, methods, classes, methodBodies);
        for (let i = 0; i < classCount; i++) {
            classes.push(this.readClass(i, constantPool, methods, classes, instances, methodBodies));
        }
        const scriptCount = this.buffer.readEncodedU30();
        const scripts = [];
        for (let i = 0; i < scriptCount; i++) {
            scripts.push(this.readScript(constantPool, methods, classes, methodBodies));
        }
        const methodBodyCount = this.buffer.readEncodedU30();
        for (let i = 0; i < methodBodyCount; i++) {
            methodBodies.push(this.readMethodBodies(methodBodyCount, methods, constantPool, classes, methodBodies));
        }
        return {
            minorVersion,
            majorVersion,
            constantPool,
            methodCount,
            methods,
            metadataCount,
            metadataInfos,
            classCount,
            instances,
            classes,
            scriptCount,
            scripts,
            methodBodyCount,
            methodBodies
        };
    }
    readConstantPool() {
        const intCount = this.buffer.readEncodedU30();
        const integers = [0];
        for (let i = 1; i < intCount; i++) {
            integers.push(this.buffer.readEncodedU32());
        }
        const uintCount = this.buffer.readEncodedU30();
        const uintegers = [0];
        for (let i = 1; i < uintCount; i++) {
            uintegers.push(this.buffer.readEncodedU32());
        }
        const doubleCount = this.buffer.readEncodedU30();
        const doubles = [NaN];
        for (let i = 1; i < doubleCount; i++) {
            doubles.push(this.buffer.readDoubleLE());
        }
        const stringCount = this.buffer.readEncodedU30();
        const strings = ["*"];
        for (let i = 1; i < stringCount; i++) {
            const size = this.buffer.readEncodedU30();
            strings.push(this.buffer.readBuffer(size).toString("utf8"));
        }
        const namespaceCount = this.buffer.readEncodedU30();
        const namespaces = [null];
        for (let i = 1; i < namespaceCount; i++) {
            const kind = this.buffer.readUInt8();
            const name = this.buffer.readEncodedU30();
            namespaces.push({
                kind,
                get name() {
                    return strings[name];
                }
            });
        }
        const nsSetCount = this.buffer.readEncodedU30();
        const nsSet = [null];
        for (let i = 1; i < nsSetCount; i++) {
            const count = this.buffer.readEncodedU30();
            const nsArray = [];
            for (let x = 0; x < count; x++) {
                const ref = this.buffer.readEncodedU30();
                nsArray.push(namespaces[ref]);
            }
            nsSet.push({ count, namespaces: nsArray });
        }
        const multinameCount = this.buffer.readEncodedU30();
        const multinames = [null];
        for (let i = 1; i < multinameCount; i++) {
            const kind = this.buffer.readUInt8();
            switch (kind) {
                case multiname_1.MultinameKind.QName:
                case multiname_1.MultinameKind.QNameA:
                    const ns = this.buffer.readEncodedU30();
                    const name = this.buffer.readEncodedU30();
                    const qname = {
                        kind,
                        get ns() {
                            return namespaces[ns];
                        },
                        get name() {
                            return strings[name];
                        }
                    };
                    multinames.push(qname);
                    break;
                case multiname_1.MultinameKind.RTQName:
                case multiname_1.MultinameKind.RTQNameA:
                    const name2 = this.buffer.readEncodedU30();
                    const rtqname = {
                        kind,
                        get name() {
                            return strings[name2];
                        }
                    };
                    multinames.push(rtqname);
                    break;
                case multiname_1.MultinameKind.RTQNAmeL:
                case multiname_1.MultinameKind.RTQNameLA:
                    const rtqnamel = { kind };
                    multinames.push(rtqnamel);
                    break;
                case multiname_1.MultinameKind.Multiname:
                case multiname_1.MultinameKind.MultinameA:
                    const name3 = this.buffer.readEncodedU30();
                    const nsSetIndex = this.buffer.readEncodedU30();
                    const multiname = {
                        kind,
                        get name() {
                            return strings[name3];
                        },
                        get nsSet() {
                            return nsSet[nsSetIndex];
                        }
                    };
                    multinames.push(multiname);
                    break;
                case multiname_1.MultinameKind.MultinameL:
                case multiname_1.MultinameKind.MultinameLA:
                    const nsSetIndex2 = this.buffer.readEncodedU30();
                    const multinamel = {
                        kind,
                        get nsSet() {
                            return nsSet[nsSetIndex2];
                        }
                    };
                    multinames.push(multinamel);
                    break;
                case multiname_1.MultinameKind.TypeName:
                    const name4 = this.buffer.readEncodedU30();
                    const paramsCount = this.buffer.readEncodedU30();
                    const params = [];
                    for (let j = 0; j < paramsCount; j++) {
                        params.push(this.buffer.readEncodedU30());
                    }
                    const typeName = {
                        kind,
                        get name() {
                            return multinames[name4];
                        },
                        get names() {
                            return params.map(p => multinames[p]);
                        }
                    };
                    multinames.push(typeName);
                    break;
                default:
                    multinames.push(null);
                    break;
            }
        }
        return {
            intCount,
            integers,
            uintCount,
            uintegers,
            doubleCount,
            doubles,
            stringCount,
            strings,
            namespaceCount,
            namespaces,
            nsSetCount,
            nsSet,
            multinameCount,
            multinames
        };
    }
    buildConstant(kind, valueIndex, constantPool) {
        switch (kind) {
            case constant_1.ConstantKind.Int:
                return {
                    kind,
                    get val() {
                        return constantPool.integers[valueIndex];
                    }
                };
            case constant_1.ConstantKind.UInt:
                return {
                    kind,
                    get val() {
                        return constantPool.uintegers[valueIndex];
                    }
                };
            case constant_1.ConstantKind.Double:
                return {
                    kind,
                    get val() {
                        return constantPool.doubles[valueIndex];
                    }
                };
            case constant_1.ConstantKind.Utf8:
                return {
                    kind,
                    get val() {
                        return constantPool.strings[valueIndex];
                    }
                };
            case constant_1.ConstantKind.PackageNamespace:
            case constant_1.ConstantKind.Namespace:
            case constant_1.ConstantKind.PackageInternalNs:
            case constant_1.ConstantKind.ProtectedNamespace:
            case constant_1.ConstantKind.ExplicitNamespace:
            case constant_1.ConstantKind.StaticProtectedNs:
            case constant_1.ConstantKind.PrivateNs:
                return {
                    kind,
                    get val() {
                        return constantPool.namespaces[valueIndex];
                    }
                };
            case constant_1.ConstantKind.True:
                return {
                    kind,
                    get val() {
                        return true;
                    }
                };
            case constant_1.ConstantKind.False:
                return {
                    kind,
                    get val() {
                        return false;
                    }
                };
            case constant_1.ConstantKind.Null:
                return {
                    kind,
                    get val() {
                        return null;
                    }
                };
            case constant_1.ConstantKind.Undefined:
                return {
                    kind,
                    get val() {
                        return undefined;
                    }
                };
        }
    }
    readMethods(methodCount, constantPool) {
        const methods = [];
        for (let i = 0; i < methodCount; i++) {
            const paramCount = this.buffer.readEncodedU30();
            const returnTypeIndex = this.buffer.readEncodedU30();
            const paramTypes = [];
            for (let y = 0; y < paramCount; y++) {
                paramTypes.push(constantPool.multinames[this.buffer.readEncodedU30()]);
            }
            const nameIndex = this.buffer.readEncodedU30();
            const flags = this.buffer.readUInt8();
            const method = {
                paramCount,
                get returnType() {
                    return constantPool.multinames[returnTypeIndex];
                },
                paramTypes,
                get name() {
                    return constantPool.strings[nameIndex];
                },
                flags
            };
            if (flags & methods_1.MethodInfoFlag.HasOptional) {
                const count = this.buffer.readEncodedU30();
                const options = [];
                for (let x = 0; x < count; x++) {
                    const valIndex = this.buffer.readEncodedU30();
                    const kind = this.buffer.readUInt8();
                    options.push(this.buildConstant(kind, valIndex, constantPool));
                }
                method.options = { count, options };
            }
            if (flags & methods_1.MethodInfoFlag.HasParamNames) {
                const names = [];
                for (let y = 0; y < paramCount; y++) {
                    const index = this.buffer.readEncodedU30();
                    names.push(constantPool.strings[index]);
                }
                method.paramNames = { names };
            }
            methods.push(method);
        }
        return methods;
    }
    readMetadataInfos(metadataCount, constantPool) {
        const metadatas = [];
        for (let i = 0; i < metadataCount; i++) {
            const nameIndex = this.buffer.readEncodedU30();
            const itemCount = this.buffer.readEncodedU30();
            const keys = [];
            for (let z = 0; z < itemCount; z++) {
                keys.push(constantPool.strings[this.buffer.readEncodedU30()]);
            }
            const values = [];
            for (let j = 0; j < itemCount; j++) {
                values.push(constantPool.strings[this.buffer.readEncodedU30()]);
            }
            metadatas.push({
                get name() {
                    return constantPool.strings[nameIndex];
                },
                itemCount,
                keys,
                values
            });
        }
        return metadatas;
    }
    readTrait(constantPool, methods, classes, methodBodies) {
        const nameIndex2 = this.buffer.readEncodedU30();
        const kindAndAttrs = this.buffer.readUInt8();
        // lower four bits
        const kind = kindAndAttrs & 15;
        // upper four bits
        const attrs = kindAndAttrs >> 4;
        let trait;
        switch (kind) {
            case trait_1.TraitKind.Slot:
            case trait_1.TraitKind.Const: {
                const slotId = this.buffer.readEncodedU30();
                const typeNameIndex = this.buffer.readEncodedU30();
                const vindex = this.buffer.readEncodedU30();
                let vkind;
                if (vindex > 0) {
                    vkind = this.buffer.readUInt8();
                }
                trait = {
                    get name() {
                        return constantPool.multinames[nameIndex2];
                    },
                    kind,
                    slotId,
                    get typeName() {
                        return constantPool.multinames[typeNameIndex];
                    },
                    value: vindex > 0
                        ? this.buildConstant(vkind, vindex, constantPool)
                        : undefined
                };
                break;
            }
            case trait_1.TraitKind.Class: {
                const slotId = this.buffer.readEncodedU30();
                const classi = this.buffer.readEncodedU30();
                trait = {
                    get name() {
                        return constantPool.multinames[nameIndex2];
                    },
                    kind,
                    get class() {
                        return classes[classi];
                    }
                };
                break;
            }
            case trait_1.TraitKind.Function: {
                const slotId = this.buffer.readEncodedU30();
                const functionIndex = this.buffer.readEncodedU30();
                trait = {
                    get name() {
                        return constantPool.multinames[nameIndex2];
                    },
                    kind,
                    slotId,
                    get func() {
                        return methods[methodIndex];
                    },
                    get funcBody() {
                        return methodBodies.find(mb => mb.methodIndex === methodIndex);
                    }
                };
                break;
            }
            case trait_1.TraitKind.Method:
            case trait_1.TraitKind.Getter:
            case trait_1.TraitKind.Setter:
                const dispId = this.buffer.readEncodedU30();
                const methodIndex = this.buffer.readEncodedU30();
                let attribute;
                if (attrs & 0x1) {
                    attribute = trait_1.TraitAttribute.Final;
                }
                else if (attrs & 0x2) {
                    attribute = trait_1.TraitAttribute.Override;
                }
                trait = {
                    attribute,
                    dispId,
                    get name() {
                        return constantPool.multinames[nameIndex2];
                    },
                    kind,
                    get method() {
                        return methods[methodIndex];
                    },
                    get methodBody() {
                        return methodBodies.find(mb => mb.methodIndex === methodIndex);
                    }
                };
                break;
        }
        if (attrs & 0x4) {
            const metadataCount = this.buffer.readEncodedU30();
            for (let j = 0; j < metadataCount; j++) {
                this.buffer.readEncodedU30();
            }
        }
        return trait;
    }
    readClass(index, constantPool, methods, classes, instances, methodBodies) {
        const cinitIndex = this.buffer.readEncodedU30();
        const traitCount = this.buffer.readEncodedU30();
        const traits = [];
        for (let i = 0; i < traitCount; i++) {
            traits.push(this.readTrait(constantPool, methods, classes, methodBodies));
        }
        return {
            get instance() {
                return instances[index];
            },
            get cinit() {
                return methods[cinitIndex];
            },
            get cinitBody() {
                return methodBodies.find(mb => mb.methodIndex === cinitIndex);
            },
            traitCount,
            traits
        };
    }
    readScript(constantPool, methods, classes, methodBodies) {
        const initIndex = this.buffer.readEncodedU30();
        const traitCount = this.buffer.readEncodedU30();
        const traits = [];
        for (let i = 0; i < traitCount; i++) {
            traits.push(this.readTrait(constantPool, methods, classes, methodBodies));
        }
        return {
            get init() {
                return methods[initIndex];
            },
            traitCount,
            traits
        };
    }
    readInstances(classCount, constantPool, methods, classes, methodBodies) {
        const instances = [];
        for (let i = 0; i < classCount; i++) {
            const nameIndex = this.buffer.readEncodedU30();
            const supernameIndex = this.buffer.readEncodedU30();
            const flags = this.buffer.readUInt8();
            let protectedNs;
            if (flags & instance_1.InstanceInfoFlag.ClassProtectedNs) {
                protectedNs = constantPool.namespaces[this.buffer.readEncodedU30()];
            }
            const interfaceCount = this.buffer.readEncodedU30();
            const interfaces = [];
            for (let z = 0; z < interfaceCount; z++) {
                interfaces.push(constantPool.multinames[this.buffer.readEncodedU30()]);
            }
            const iinitIndex = this.buffer.readEncodedU30();
            const traitCount = this.buffer.readEncodedU30();
            const traits = [];
            for (let x = 0; x < traitCount; x++) {
                traits.push(this.readTrait(constantPool, methods, classes, methodBodies));
            }
            instances.push({
                get class() {
                    return classes[i];
                },
                get name() {
                    return constantPool.multinames[nameIndex];
                },
                get supername() {
                    return constantPool.multinames[supernameIndex];
                },
                flags,
                protectedNs,
                interfaceCount,
                interfaces,
                get iinit() {
                    return methods[iinitIndex];
                },
                traitCount,
                traits
            });
        }
        return instances;
    }
    readExceptions(exceptionCount, multinames) {
        const exceptions = [];
        for (let i = 0; i < exceptionCount; i++) {
            const from = this.buffer.readEncodedU30();
            const to = this.buffer.readEncodedU30();
            const target = this.buffer.readEncodedU30();
            const excTypeIndex = this.buffer.readEncodedU30();
            const varNameIndex = this.buffer.readEncodedU30();
            exceptions.push({
                from,
                to,
                target,
                get excType() {
                    return multinames[excTypeIndex];
                },
                get varName() {
                    return multinames[varNameIndex];
                }
            });
        }
        return exceptions;
    }
    readMethodBodies(methodBodyCount, methods, constantPool, classes, methodBodies) {
        const methodIndex = this.buffer.readEncodedU30();
        const maxStack = this.buffer.readEncodedU30();
        const localCount = this.buffer.readEncodedU30();
        const initScopeDepth = this.buffer.readEncodedU30();
        const maxScopeDepth = this.buffer.readEncodedU30();
        const codeLength = this.buffer.readEncodedU30();
        const startOffset = this.buffer.readOffset;
        const code = [];
        while (this.buffer.readOffset < startOffset + codeLength) {
            const byteOffset = this.buffer.readOffset - startOffset;
            code.push(Object.assign(this.readInstruction(constantPool), { byteOffset }));
        }
        const exceptionCount = this.buffer.readEncodedU30();
        const exceptions = this.readExceptions(exceptionCount, constantPool.multinames);
        const traitCount = this.buffer.readEncodedU30();
        const traits = [];
        for (let y = 0; y < traitCount; y++) {
            traits.push(this.readTrait(constantPool, methods, classes, methodBodies));
        }
        return {
            methodIndex,
            get method() {
                return methods[methodIndex];
            },
            maxStack,
            localCount,
            initScopeDepth,
            maxScopeDepth,
            codeLength,
            code,
            exceptionCount,
            exceptions,
            traitCount,
            traits
        };
    }
    readInstruction(constantPool) {
        const code = this.buffer.readUInt8();
        switch (code) {
            case bytecode_1.InstructionCode.add: {
                return { code };
            }
            case bytecode_1.InstructionCode.add_i: {
                return { code };
            }
            case bytecode_1.InstructionCode.astype: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.astypelate: {
                return { code };
            }
            case bytecode_1.InstructionCode.bitand: {
                return { code };
            }
            case bytecode_1.InstructionCode.bitnot: {
                return { code };
            }
            case bytecode_1.InstructionCode.bitor: {
                return { code };
            }
            case bytecode_1.InstructionCode.bitxor: {
                return { code };
            }
            case bytecode_1.InstructionCode.call: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.callmethod: {
                const operand0 = this.buffer.readEncodedU30();
                const operand1 = this.buffer.readEncodedU30();
                return { code, operand0, operand1 };
            }
            case bytecode_1.InstructionCode.callproperty: {
                const index = this.buffer.readEncodedU30();
                const argCount = this.buffer.readEncodedU30();
                return {
                    argCount,
                    code,
                    get name() {
                        return constantPool.multinames[index];
                    }
                };
            }
            case bytecode_1.InstructionCode.callproplex: {
                const operand0 = this.buffer.readEncodedU30();
                const operand1 = this.buffer.readEncodedU30();
                return { code, operand0, operand1 };
            }
            case bytecode_1.InstructionCode.callpropvoid: {
                const index = this.buffer.readEncodedU30();
                const argCount = this.buffer.readEncodedU30();
                return {
                    argCount,
                    code,
                    get name() {
                        return constantPool.multinames[index];
                    }
                };
            }
            case bytecode_1.InstructionCode.callstatic: {
                const operand0 = this.buffer.readEncodedU30();
                const operand1 = this.buffer.readEncodedU30();
                return { code, operand0, operand1 };
            }
            case bytecode_1.InstructionCode.callsuper: {
                const operand0 = this.buffer.readEncodedU30();
                const operand1 = this.buffer.readEncodedU30();
                return { code, operand0, operand1 };
            }
            case bytecode_1.InstructionCode.callsupervoid: {
                const operand0 = this.buffer.readEncodedU30();
                const operand1 = this.buffer.readEncodedU30();
                return { code, operand0, operand1 };
            }
            case bytecode_1.InstructionCode.checkfilter: {
                return { code };
            }
            case bytecode_1.InstructionCode.coerce: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.coerce_a: {
                return { code };
            }
            case bytecode_1.InstructionCode.coerce_s: {
                return { code };
            }
            case bytecode_1.InstructionCode.construct: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.contructprop: {
                const index = this.buffer.readEncodedU30();
                const argCount = this.buffer.readEncodedU30();
                return {
                    argCount,
                    code,
                    get name() {
                        return constantPool.multinames[index];
                    }
                };
            }
            case bytecode_1.InstructionCode.constructsuper: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.convert_b: {
                return { code };
            }
            case bytecode_1.InstructionCode.convert_i: {
                return { code };
            }
            case bytecode_1.InstructionCode.convert_d: {
                return { code };
            }
            case bytecode_1.InstructionCode.convert_o: {
                return { code };
            }
            case bytecode_1.InstructionCode.convert_u: {
                return { code };
            }
            case bytecode_1.InstructionCode.convert_s: {
                return { code };
            }
            case bytecode_1.InstructionCode.debug: {
                const operand0 = this.buffer.readUInt8();
                const operand1 = this.buffer.readEncodedU30();
                const operand2 = this.buffer.readUInt8();
                const operand3 = this.buffer.readEncodedU30();
                return { code, operand0, operand1, operand2, operand3 };
            }
            case bytecode_1.InstructionCode.debugfile: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.debugline: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.declocal: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.declocal_i: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.decrement: {
                return { code };
            }
            case bytecode_1.InstructionCode.decrement_i: {
                return { code };
            }
            case bytecode_1.InstructionCode.deleteproperty: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.divide: {
                return { code };
            }
            case bytecode_1.InstructionCode.dup: {
                return { code };
            }
            case bytecode_1.InstructionCode.dxns: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.dxnslate: {
                return { code };
            }
            case bytecode_1.InstructionCode.equals: {
                return { code };
            }
            case bytecode_1.InstructionCode.esc_xattr: {
                return { code };
            }
            case bytecode_1.InstructionCode.esc_xelem: {
                return { code };
            }
            case bytecode_1.InstructionCode.findproperty: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.findpropstrict: {
                const index = this.buffer.readEncodedU30();
                return {
                    code,
                    get name() {
                        return constantPool.multinames[index];
                    }
                };
            }
            case bytecode_1.InstructionCode.getdescendants: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.getglobalscope: {
                return { code };
            }
            case bytecode_1.InstructionCode.getglobalslot: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.getlex: {
                const index = this.buffer.readEncodedU30();
                return {
                    code,
                    get name() {
                        return constantPool.multinames[index];
                    }
                };
            }
            case bytecode_1.InstructionCode.getlocal: {
                const index = this.buffer.readEncodedU30();
                return { code, index };
            }
            case bytecode_1.InstructionCode.getlocal_0: {
                return { code };
            }
            case bytecode_1.InstructionCode.getlocal_1: {
                return { code };
            }
            case bytecode_1.InstructionCode.getlocal_2: {
                return { code };
            }
            case bytecode_1.InstructionCode.getlocal_3: {
                return { code };
            }
            case bytecode_1.InstructionCode.getproperty: {
                const index = this.buffer.readEncodedU30();
                return {
                    code,
                    get name() {
                        return constantPool.multinames[index];
                    }
                };
            }
            case bytecode_1.InstructionCode.getscopeobject: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.getslot: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.getsuper: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.greaterequals: {
                return { code };
            }
            case bytecode_1.InstructionCode.greaterthan: {
                return { code };
            }
            case bytecode_1.InstructionCode.hasnext: {
                return { code };
            }
            case bytecode_1.InstructionCode.hasnext2: {
                const operand0 = this.buffer.readEncodedU30();
                const operand1 = this.buffer.readEncodedU30();
                return { code, operand0, operand1 };
            }
            case bytecode_1.InstructionCode.ifeq: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.iffalse: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.ifge: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.ifgt: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.ifle: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.iflt: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.ifnge: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.ifngt: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.ifnle: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.ifnlt: {
                const offset = this.buffer.readS24();
                return { code, offset };
            }
            case bytecode_1.InstructionCode.ifne: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.ifstricteq: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.ifstrictne: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.iftrue: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.in: {
                return { code };
            }
            case bytecode_1.InstructionCode.inclocal: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.inclocal_i: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.increment: {
                return { code };
            }
            case bytecode_1.InstructionCode.increment_i: {
                return { code };
            }
            case bytecode_1.InstructionCode.initproperty: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.instanceof: {
                return { code };
            }
            case bytecode_1.InstructionCode.istype: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.istypelate: {
                return { code };
            }
            case bytecode_1.InstructionCode.jump: {
                const operand0 = this.buffer.readS24();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.kill: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.label: {
                return { code };
            }
            case bytecode_1.InstructionCode.lessequals: {
                return { code };
            }
            case bytecode_1.InstructionCode.lessthan: {
                return { code };
            }
            case bytecode_1.InstructionCode.lookupswitch: {
                const operand0 = this.buffer.readS24();
                const count = this.buffer.readEncodedU30();
                const cases = [];
                for (let i = 0; i < count + 1; i++) {
                    cases.push(this.buffer.readS24());
                }
                return { code, operand0, cases };
            }
            case bytecode_1.InstructionCode.lshift: {
                return { code };
            }
            case bytecode_1.InstructionCode.modulo: {
                return { code };
            }
            case bytecode_1.InstructionCode.multiply: {
                return { code };
            }
            case bytecode_1.InstructionCode.multiply_i: {
                return { code };
            }
            case bytecode_1.InstructionCode.negate: {
                return { code };
            }
            case bytecode_1.InstructionCode.negate_i: {
                return { code };
            }
            case bytecode_1.InstructionCode.newactivation: {
                return { code };
            }
            case bytecode_1.InstructionCode.newarray: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.newcatch: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.newclass: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.newfunction: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.newobject: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.nextname: {
                return { code };
            }
            case bytecode_1.InstructionCode.nextvalue: {
                return { code };
            }
            case bytecode_1.InstructionCode.nop: {
                return { code };
            }
            case bytecode_1.InstructionCode.not: {
                return { code };
            }
            case bytecode_1.InstructionCode.pop: {
                return { code };
            }
            case bytecode_1.InstructionCode.popscope: {
                return { code };
            }
            case bytecode_1.InstructionCode.pushbyte: {
                const byteValue = this.buffer.readUInt8();
                return { code, byteValue };
            }
            case bytecode_1.InstructionCode.pushdouble: {
                const index = this.buffer.readEncodedU30();
                return {
                    code,
                    get value() {
                        return constantPool.doubles[index];
                    }
                };
            }
            case bytecode_1.InstructionCode.pushfalse: {
                return { code };
            }
            case bytecode_1.InstructionCode.pushint: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.pushnamespace: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.pushnan: {
                return { code };
            }
            case bytecode_1.InstructionCode.pushnull: {
                return { code };
            }
            case bytecode_1.InstructionCode.pushscope: {
                return { code };
            }
            case bytecode_1.InstructionCode.pushshort: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.pushstring: {
                const index = this.buffer.readEncodedU30();
                return {
                    code,
                    get value() {
                        return constantPool.strings[index];
                    }
                };
            }
            case bytecode_1.InstructionCode.pushtrue: {
                return { code };
            }
            case bytecode_1.InstructionCode.pushuint: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.undefined: {
                return { code };
            }
            case bytecode_1.InstructionCode.pushwith: {
                return { code };
            }
            case bytecode_1.InstructionCode.returnvalue: {
                return { code };
            }
            case bytecode_1.InstructionCode.returnvoid: {
                return { code };
            }
            case bytecode_1.InstructionCode.rshift: {
                return { code };
            }
            case bytecode_1.InstructionCode.setlocal: {
                const index = this.buffer.readEncodedU30();
                return { code, index };
            }
            case bytecode_1.InstructionCode.setlocal_0: {
                return { code };
            }
            case bytecode_1.InstructionCode.setlocal_1: {
                return { code };
            }
            case bytecode_1.InstructionCode.setlocal_2: {
                return { code };
            }
            case bytecode_1.InstructionCode.setlocal_3: {
                return { code };
            }
            case bytecode_1.InstructionCode.setglobalslot: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.setproperty: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.setslot: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.setsuper: {
                const operand0 = this.buffer.readEncodedU30();
                return { code, operand0 };
            }
            case bytecode_1.InstructionCode.strictequals: {
                return { code };
            }
            case bytecode_1.InstructionCode.subtract: {
                return { code };
            }
            case bytecode_1.InstructionCode.subtract_i: {
                return { code };
            }
            case bytecode_1.InstructionCode.swap: {
                return { code };
            }
            case bytecode_1.InstructionCode.throw: {
                return { code };
            }
            case bytecode_1.InstructionCode.typeof: {
                return { code };
            }
            case bytecode_1.InstructionCode.urshift: {
                return { code };
            }
            case bytecode_1.InstructionCode.applytype: {
                return { code };
            }
            case bytecode_1.InstructionCode.bkpt: {
                return { code };
            }
            case bytecode_1.InstructionCode.lf32: {
                return { code };
            }
            case bytecode_1.InstructionCode.lf64: {
                return { code };
            }
            case bytecode_1.InstructionCode.li16: {
                return { code };
            }
            case bytecode_1.InstructionCode.li32: {
                return { code };
            }
            case bytecode_1.InstructionCode.li8: {
                return { code };
            }
            case bytecode_1.InstructionCode.sf32: {
                return { code };
            }
            case bytecode_1.InstructionCode.sf64: {
                return { code };
            }
            case bytecode_1.InstructionCode.si16: {
                return { code };
            }
            case bytecode_1.InstructionCode.si32: {
                return { code };
            }
            case bytecode_1.InstructionCode.si8: {
                return { code };
            }
            case bytecode_1.InstructionCode.sxi1: {
                return { code };
            }
            case bytecode_1.InstructionCode.sxi16: {
                return { code };
            }
            case bytecode_1.InstructionCode.sxi8: {
                return { code };
            }
            default: {
                throw new Error("could not read instruction with code " + code);
            }
        }
    }
}
exports.default = AbcFileReader;
